# Internode filters
DEFAULT_FILTER = '0'
WITNESS_MASTERNODE_FILTER = 'wit_mn'
DELEGATE_DELEGATE_FILTER = 'del_del'
MASTERNODE_DELEGATE_FILTER = 'mn_del'
WITNESS_DELEGATE_FILTER = 'wit_del'
TRANSACTION_FILTER = 'tx'

# Catchup Filters
BLOCK_IDX_REQ_FILTER = 'blk_idx_req'

# New Block Notif Filters
NEW_BLK_NOTIF_FILTER = 'new_blk_notif'

# IPC Filters
STATESYNC_FILTER = 'state_sync'
