# constants for transaction batcher

# in secs
BATCHER_SLEEP_INTERVAL = 1
MAX_TXN_SUBMISSION_DELAY = 5

# TODO determine optimum size based on communication and storage efficiencies
MAX_TXNS_PER_SUB_BLOCK = 100

