from unittest import TestCase
from unittest.mock import patch, MagicMock

from cilantro_ee.core.containers.merkle_tree import MerkleTree
from cilantro_ee.constants.system_config import *
from cilantro_ee.nodes.masternode.block_contender import BlockContender


# TODO write these
class TestBlockContender(TestCase):
    pass

